from __future__ import annotations
from pathlib import Path
from shutil import rmtree
from typing import Iterator
from xml.etree.ElementTree import Element

from ...xml import encode
from ..data import Layout
from ..sequence import decode_layout
from ..chapter import list_chapter_files
from ..utils import xml_files, read_xml_file, search_xml_children, XML_Info
from .extraction import extract_footnote_references
from .mark import transform2mark, search_marks, Mark


def append_footnote_for_chapters(chapter_path: Path, footnote_path: Path, output_path: Path) -> None:
  if output_path.exists():
    rmtree(output_path)

  output_path.mkdir(parents=True, exist_ok=True)
  footnote_reader = _FootnoteReader(footnote_path)

  for chapter_id, file_path in list_chapter_files(chapter_path):
    chapter_element = Element("chapter")
    chapter_footnotes: list[tuple[int, int, Element, Element]] = []

    for raw_layout_element in read_xml_file(file_path):
      layout_element: Layout = raw_layout_element
      layout = decode_layout(raw_layout_element)
      footnote_page = footnote_reader.read(layout.page_index)
      if footnote_page is not None:
        layout_element, layout_footnotes = _parse_layout_and_mark(
          footnote_page=footnote_page,
          layout=layout,
        )
        for id, mark_element, footnote_element in layout_footnotes:
          chapter_footnotes.append((layout.page_index, id, mark_element, footnote_element))
      chapter_element.append(layout_element)

    next_mark_id: int = 1
    for _, _, mark_element, footnote_element in sorted(
      chapter_footnotes,
      key=lambda x: (x[0], x[1]),
    ):
      footnote_element.set("id", str(next_mark_id))
      mark_element.set("id", str(next_mark_id))
      chapter_element.append(footnote_element)
      next_mark_id += 1

    if chapter_id is None:
      file_path = output_path / "chapter.xml"
    else:
      file_path = output_path / f"chapter_{chapter_id}.xml"

    with open(file_path, "w", encoding="utf-8") as file:
      file.write(encode(chapter_element))

def _parse_layout_and_mark(footnote_page: _FootnotePage, layout: Layout):
  layout_element = layout.to_xml()
  footnotes: list[tuple[int, Element, Element]] = []

  for line_element, _ in search_xml_children(layout_element):
    if line_element.tag != "line":
      continue
    raw_text = line_element.text.strip()
    line_element.text = None
    last_mark_element: Element | None = None

    for cell in search_marks(raw_text):
      is_footnote_mark = False
      if isinstance(cell, Mark):
        footnote_element = footnote_page.pop(cell)
        if footnote_element is not None:
          footnote_id = int(footnote_element.get("id", "-1"))
          mark_element = Element("mark")
          mark_element.text = str(cell)
          line_element.append(mark_element)
          footnotes.append((footnote_id, mark_element, footnote_element))
          is_footnote_mark = True
          last_mark_element = mark_element

      if not is_footnote_mark:
        if last_mark_element is None:
          if line_element.text:
            line_element.text += str(cell)
          else:
            line_element.text = str(cell)
        elif last_mark_element.tail:
          last_mark_element.tail += str(cell)
        else:
          last_mark_element.tail = str(cell)

      # TODO: footnote_page 如果剩余 mark 没有匹配应该交给 LLM 处理

  return layout_element, footnotes

class _FootnoteReader:
  def __init__(self, footnote_path: Path):
    self._infos_iter: Iterator[XML_Info] = iter(xml_files(footnote_path))
    self._buffer_info: XML_Info | None = None
    self.buffer_page: _FootnotePage | None = None

  def read(self, page_index: int) -> _FootnotePage | None:
    while True:
      if self._buffer_info is None:
        self._forward_buffer_info(page_index)
        if self._buffer_info is None:
          return None # read to the end

      file_path, _, file_page_index, _ = self._buffer_info
      if file_page_index < page_index:
        self._forward_buffer_info(page_index)
      elif file_page_index > page_index:
        return None
      else:
        if self.buffer_page is None:
          self.buffer_page = _FootnotePage(read_xml_file(file_path))
        return self.buffer_page

  def _forward_buffer_info(self, page_index: int):
    while True:
      info = next(self._infos_iter, None)
      if info is None:
        self._buffer_info = None
        self.buffer_page = None
        break
      _, _, file_page_index, _ = info
      if file_page_index >= page_index:
        self._buffer_info = info
        self.buffer_page = None
        break

class _FootnotePage:
  def __init__(self, element: Element):
    self._mark2element: dict[Mark, Element] = {}
    for footnote in element:
      if footnote.tag != "footnote":
        continue
      mark_element = footnote.find("mark")
      if mark_element is None:
        continue
      mark = transform2mark(mark_element.text.strip())
      if mark is None:
        continue
      self._mark2element[mark] = footnote

  def pop(self, mark: Mark) -> Element | None:
    return self._mark2element.pop(mark, None)

def generate_footnote_references(sequence_path: Path, output_path: Path) -> None:
  if output_path.exists():
    rmtree(output_path)
  output_path.mkdir(parents=True, exist_ok=True)

  for page_index, page_element in _extract_page_element(sequence_path):
    file_path = output_path / f"page_{page_index}.xml"
    with open(file_path, "w", encoding="utf-8") as file:
      file.write(encode(page_element))

def _extract_page_element(sequence_path: Path):
  page_element: Element | None = None
  page_index: int = -1
  next_footnote_id: int = -1

  # TODO: 检查每页的内容形式上是否完整，若有问题，交给 LLM 解决
  for mark_page_index, mark, layouts in extract_footnote_references(sequence_path):
    if page_element is None or mark_page_index != page_index:
      if page_element is not None:
        yield page_index, page_element
      page_element = Element("page")
      page_index = mark_page_index
      next_footnote_id = 1
      page_element.set("page-index", str(page_index))

    footnote_element = Element("footnote")
    footnote_element.set("id", str(next_footnote_id))
    mark_element = Element("mark")
    mark_element.text = mark.char
    footnote_element.append(mark_element)
    for layout in layouts:
      footnote_element.append(layout.to_xml())

    page_element.append(footnote_element)
    next_footnote_id += 1

  if page_element is not None:
    yield page_index, page_element


from io import StringIO
from typing import Callable
from html import escape as escape_html
from xml.etree.ElementTree import Element

from .tag import Tag, TagKind
from .parser import parse_tags
from .transform import element_to_tag

# why implement XML encoding?
# https://github.com/oomol-lab/pdf-craft/issues/149
def encode_friendly(element: Element, indent: int = 2) -> str:
  buffer = StringIO()
  _encode_element(
    buffer=buffer,
    element=element,
    indent=indent,
    depth=0,
    escape=_escape_text,
  )
  return buffer.getvalue()

def _escape_text(text: str) -> str:
  buffer = StringIO()
  for cell in parse_tags(text):
    if isinstance(cell, Tag):
      cell = escape_html(str(cell))
    buffer.write(cell)
  return buffer.getvalue()

def encode(element: Element, indent: int = 2) -> str:
  buffer = StringIO()
  _encode_element(
    buffer=buffer,
    element=element,
    indent=indent,
    depth=0,
    escape=escape_html,
  )
  return buffer.getvalue()

_TINY_TEXT_LEN = 35

def _encode_element(
      buffer: StringIO,
      element: Element,
      indent: int,
      depth: int,
      escape: Callable[[str], str],
    ) -> None:

  _write_indent(buffer, indent, depth)
  if len(element) == 0 and not element.text:
    tag = element_to_tag(element, TagKind.SELF_CLOSING)
    buffer.write(str(tag))
  else:
    text = (element.text or "").strip()
    opening_tag = element_to_tag(element, TagKind.OPENING)
    closing_tag = element_to_tag(element, TagKind.CLOSING)
    buffer.write(str(opening_tag))
    is_one_line = (
      len(text) <= _TINY_TEXT_LEN and
      len(element) == 0 and
      "\n" not in text
    )
    if text:
      if not is_one_line:
        buffer.write("\n")
        _write_indent(buffer, indent, depth + 1)
      buffer.write(escape(text))

    for child in element:
      buffer.write("\n")
      _encode_element(
        buffer=buffer,
        element=child,
        indent=indent,
        depth=depth + 1,
        escape=escape,
      )
      child_tail = (child.tail or "").strip()
      if child_tail:
        buffer.write("\n")
        _write_indent(buffer, indent, depth + 1)
        buffer.write(escape(child_tail))

    if not is_one_line:
      buffer.write("\n")
      _write_indent(buffer, indent, depth)

    buffer.write(str(closing_tag))

def _write_indent(buffer: StringIO, indent: int, depth: int) -> None:
  for _ in range(indent * depth):
    buffer.write(" ")


from typing import Protocol
from PIL.Image import Image
from doc_page_extractor import TableLayoutParsedFormat, ExtractedResult


class DocExtractorProtocol(Protocol):
  def prepare_models(self) -> None:
    raise NotImplementedError("must implement prepare_models method")

  def extract(
    self,
    image: Image,
    extract_formula: bool,
    extract_table_format: TableLayoutParsedFormat | None = None,
    ocr_for_each_layouts: bool = False,
    adjust_points: bool = False
  ) -> ExtractedResult:
    raise NotImplementedError("must implement extract method")


import sys

from pathlib import Path
from typing import Iterable, Generator
from xml.etree.ElementTree import Element

from .common import State, Phase, SequenceType, Truncation
from .request import SequenceRequest, RawPage
from ...llm import LLM
from ...xml import encode
from ..data import ASSET_LAYOUT_KINDS
from ..utils import (
  remove_file,
  read_xml_file,
  xml_files,
  search_xml_children,
  Context,
  MultiThreads,
  Partition,
  PartitionTask,
)


def extract_ocr(llm: LLM, context: Context[State], ocr_path: Path, threads: MultiThreads) -> None:
  sequence = _Sequence(llm, context, threads)
  return sequence.to_sequences(ocr_path)

_LayoutLines = dict[int, tuple[Element, Element]]

class _Sequence:
  def __init__(self, llm: LLM, context: Context[State], threads: MultiThreads) -> None:
    self._llm: LLM = llm
    self._ctx: Context[State] = context
    self._threads: MultiThreads = threads

  def to_sequences(self, ocr_path: Path):
    save_path = self._ctx.path.joinpath(Phase.EXTRACTION.value)
    save_path.mkdir(parents=True, exist_ok=True)

    self._ctx.reporter.set(max_count=len(xml_files(ocr_path)))
    def _add_progress_by_pages(begin, end):
      page_count = end[0] - begin[0] + 1
      self._ctx.reporter.increment(page_count)

    partition: Partition[tuple[int], State, SequenceRequest] = Partition(
      dimension=1,
      context=self._ctx,
      sequence=((r.begin, r.end, r) for r in self._split_requests(ocr_path)),
      done=_add_progress_by_pages,
      remove=lambda begin, end: remove_file(save_path / f"pages_{begin[0]}_{end[0]}.xml"),
    )
    with partition:
      self._threads.run(
        next_task=partition.pop_task,
        thread_payload=lambda: None,
        invoke=lambda _, task: self._emit_request(save_path, task),
      )

  def _split_requests(self, ocr_path: Path) -> Generator[SequenceRequest, None, None]:
    max_request_data_tokens = self._ctx.state["max_request_data_tokens"]
    request = SequenceRequest()
    request_tokens: int = 0

    for xml_path, _, page_index, _ in xml_files(ocr_path):
      raw_page = RawPage(
        raw_element=read_xml_file(xml_path),
        page_index=page_index,
      )
      if not raw_page.children: # empty page
        continue

      tokens = raw_page.tokens_count(self._llm)
      if request_tokens > 0 and request_tokens + tokens > max_request_data_tokens:
        yield request
        request = SequenceRequest()

      request.append(page_index, raw_page)
      request_tokens += tokens

    if request_tokens > 0:
      yield request

  def _emit_request(self, save_path: Path, task: PartitionTask[tuple[int], State, SequenceRequest]):
    with task:
      begin = task.begin[0]
      end = task.end[0]
      request = task.payload
      request_xml = request.inject_ids_and_get_xml()
      resp_xml = self._request_sequences(request_xml)
      data_xml = Element("pages")
      data_xml.set("begin-page-index", str(begin))
      data_xml.set("end-page-index", str(end))

      for page in self._gen_pages_with_sequences(
        request=request,
        raw_page_xmls=request_xml,
        resp_xml=resp_xml,
      ):
        data_xml.append(page)

      data_file_path = save_path / f"pages_{begin}_{end}.xml"
      with open(data_file_path, mode="w", encoding="utf-8") as file:
        file.write(encode(data_xml))

  def _request_sequences(self, request_xml: Element) -> Element:
    next_id: int = 1
    for page in request_xml:
      for layout in page:
        for child, _ in search_xml_children(layout):
          child.set("id", str(next_id))
          next_id += 1

    return self._llm.request_xml(
      template_name="sequence/extraction",
      user_data=request_xml,
    )

  def _gen_pages_with_sequences(
        self,
        request: SequenceRequest,
        raw_page_xmls: Iterable[Element],
        resp_xml: Element,
      ) -> Generator[Element, None, None]:

    raw_pages: dict[int, Element] = {}
    resp_pages: dict[int, Element] = {}

    for pages, page_xmls in ((raw_pages, raw_page_xmls), (resp_pages, resp_xml)):
      for page_xml in page_xmls:
        try:
          page_index = int(page_xml.get("page-index", None))
          pages[page_index] = page_xml
        except (ValueError, TypeError):
          pass

    for page_index in sorted(list(raw_pages.keys())):
      raw_page = raw_pages[page_index]
      resp_page = resp_pages.get(page_index, None)
      if resp_page is None:
        continue
      yield self._create_page_with_sequences(
        request=request,
        page_index=page_index,
        raw_page_element=raw_page,
        resp_page=resp_page,
      )

  def _create_page_with_sequences(
        self,
        request: SequenceRequest,
        page_index: int,
        raw_page_element: Element,
        resp_page: Element,
      ) -> Element:

    layout_lines: _LayoutLines = {}
    new_page = Element(
      "page",
      self._pick_attrib(resp_page, ("page-index", "type")),
    )
    for layout in raw_page_element:
      for line in layout:
        try:
          id: int = int(line.get("id", None))
          layout_lines[id] = (layout, line)
        except (ValueError, TypeError):
          pass

    text, footnote = self._text_and_footnote_groups(resp_page, layout_lines)
    begin: tuple[int, int, Element] | None = None
    end: tuple[int, int, Element] | None = None
    raw_page = request.raw_page(page_index)

    for group_pair in (text, footnote):
      if group_pair is None:
        continue
      group, ids = group_pair
      list_ids = list(set(ids))
      sequence = self._create_sequence(
        layout_lines=layout_lines,
        group=group,
        group_ids=list_ids,
        raw_page=raw_page,
      )
      new_page.append(sequence)
      end = self._term_sequence(sequence, list_ids)
      if begin is None:
        begin = self._term_sequence(sequence, list_ids)

    if raw_page is not None:
      if begin:
        begin_line_id, _, sequence = begin
        for element in reversed(list(raw_page.assets_in_range(
          before_line_id=begin_line_id,
        ))):
          sequence.insert(0, element)

      if begin and end:
        _, after_line_id, _ = begin
        begin_line_id, _, _ = end
        if after_line_id < begin_line_id:
          sequence.extend(raw_page.assets_in_range(
            after_line_id=after_line_id,
            before_line_id=begin_line_id,
          ))

      if end:
        _, end_line_id, sequence = end
        for element in raw_page.assets_in_range(
          after_line_id=end_line_id,
        ):
          sequence.append(element)

    for child, _ in search_xml_children(new_page):
      if child.tag == "line":
        child.attrib.pop("id", None)

    return new_page

  def _text_and_footnote_groups(self, resp_page: Element, layout_lines: _LayoutLines):
    text: tuple[Element, list[int]] | None = None
    footnote: tuple[Element, list[int]] | None = None
    origin_text_ids: set[int] = set()
    origin_footnote_ids: set[int] = set()

    for group in resp_page:
      type = group.get("type", None)
      if type == SequenceType.TEXT:
        text_ids = self._ids_from_group(group)
        if text_ids:
          text = (group, text_ids)
          origin_text_ids.update(text_ids)
      elif type == SequenceType.FOOTNOTE:
        footnote_ids = self._ids_from_group(group)
        if footnote_ids:
          footnote = (group, footnote_ids)
          origin_footnote_ids.update(footnote_ids)

    if text and footnote:
      text_group, text_ids = text
      footnote_group, footnote_ids = footnote

      # There may be overlap between the two.
      # In this case, the footnote header should be used as the reference for re-cutting.
      if text_ids[-1] >= footnote_ids[0]:
        text_id_cut_index: int = -1
        for i, id in enumerate(text_ids):
          if id >= footnote_ids[0]:
            text_id_cut_index = i
            break
        text = (text_group, text_ids[:text_id_cut_index])
        footnote_ids = sorted(footnote_ids + text_ids[text_id_cut_index:])
        footnote = (footnote_group, footnote_ids)
        if footnote_ids[-1] not in origin_footnote_ids:
          footnote_group.set(
            "truncation-end",
            text_group.get("truncation-end", None),
          )
          origin_footnote_ids = set(footnote_ids)

    if text:
      text_group, text_ids = text
      before_id: int | None = None
      if footnote:
        _, footnote_ids = footnote
        if footnote_ids:
          before_id = footnote_ids[0]

      text_ids = self._fill_gap_with_lines(
        raw_ids=text_ids,
        layout_lines=layout_lines,
        before_id=before_id,
        ignore_origin_abandon=True,
      )
      if text_ids:
        text = (text_group, text_ids)
        if text_ids[-1] not in origin_text_ids:
          text_group.set("truncation-end", Truncation.UNCERTAIN.value)
      else:
        text = None

    if footnote:
      footnote_group, footnote_ids = footnote
      footnote_ids = text_ids = self._fill_gap_with_lines(
        raw_ids=footnote_ids,
        layout_lines=layout_lines,
        before_id=None,
        ignore_origin_abandon=False,
      )
      if footnote_ids:
        footnote = (footnote_group, footnote_ids)
        if footnote_ids[-1] not in origin_footnote_ids:
          footnote_group.set("truncation-end", Truncation.UNCERTAIN.value)
      else:
        footnote = None

    return text, footnote

  def _ids_from_group(self, group: Element) -> list[int]:
    ids: list[int] = []
    for resp_line in group:
      for id in self._iter_line_ids(resp_line):
        ids.append(id)
    return sorted(ids)

  def _fill_gap_with_lines(
        self,
        raw_ids: list[int],
        layout_lines: _LayoutLines,
        before_id: int | None,
        ignore_origin_abandon: bool,
      ) -> list[int]:

    target_ids: list[int] = []
    if not raw_ids:
      return target_ids

    def check_and_collect(gap_id: int):
      gap = layout_lines.get(gap_id, None)
      if gap is None:
        return
      layout, _ = gap
      if layout.tag == "abandon" or layout.tag in ASSET_LAYOUT_KINDS:
        return
      target_ids.append(gap_id)

    pre_id = raw_ids[0] - 1
    for id in raw_ids:
      for gap_id in range(pre_id + 1, id):
        check_and_collect(gap_id)

      pre_id = id
      origin = layout_lines.get(id, None)
      if origin is not None:
        layout, _ = origin
        if not ignore_origin_abandon or layout.tag != "abandon":
          target_ids.append(id)

    if before_id is not None:
      for gap_id in range(pre_id + 1, before_id):
        check_and_collect(gap_id)

    return target_ids

  def _term_sequence(self, sequence: Element, ids: Iterable[int]) -> tuple[int, int, Element]:
    begin_id: int = sys.maxsize
    end_id: int = -1
    for id in ids:
      if id < begin_id:
        begin_id = id
      if id > end_id:
        end_id = id
    return begin_id, end_id, sequence

  def _create_sequence(
        self,
        layout_lines: _LayoutLines,
        group: Element,
        group_ids: list[int],
        raw_page: RawPage | None,
      ) -> Generator[Element, None, None]:

    # TODO: 由于 asset 的插入，会导致“断裂分析”结果不一定准确，此处需要重新设计
    group_id_and_asset_element: Iterable[tuple[int, Element | None]] = ()
    if raw_page is not None:
      group_id_and_asset_element = raw_page.inject_assets(group_ids)
    else:
      group_id_and_asset_element = ((id, None) for id in sorted(group_ids))

    current_layout: tuple[Element, Element] | None = None
    sequence = Element(
      "sequence",
      self._pick_attrib(
        element=group,
        keys=("type", "truncation-begin", "truncation-end"),
      ),
    )
    for id, asset_layout in group_id_and_asset_element:
      if asset_layout is not None:
        sequence.append(asset_layout)
        continue

      result = layout_lines.get(id, None)
      if result is None:
        continue

      layout, line = result
      if current_layout is not None:
        raw_layout, new_layout = current_layout
        if raw_layout != layout:
          sequence.append(new_layout)
          current_layout = None

      if current_layout is None:
        new_layout = Element(
          layout.tag,
          self._reject_attrib(
            element=layout,
            keys=("indent", "touch-end"),
          ),
        )
        current_layout = (layout, new_layout)

      _, new_layout = current_layout
      new_line = Element(line.tag, line.attrib)
      new_line.text = line.text
      new_line.tail = line.tail
      new_layout.append(new_line)

    if current_layout is not None:
      _, new_layout = current_layout
      sequence.append(new_layout)

    return sequence

  def _pick_attrib(self, element: Element, keys: tuple[str, ...]) -> dict[str, str]:
    attr: dict[str, str] = {}
    for key in keys:
      value = element.get(key, None)
      if value is not None:
        attr[key] = value
    return attr

  def _reject_attrib(self, element: Element, keys: tuple[str, ...]) -> dict[str, str]:
    attr: dict[str, str] = {}
    for key, value in element.attrib.items():
      if key not in keys:
        attr[key] = value
    return attr

  def _iter_line_ids(self, line: Element) -> Generator[int, None, None]:
    ids = line.get("id", None)
    if ids is None:
      return

    ids = ids.split("-")
    id_begin: int
    id_end: int

    if len(ids) == 1:
      id_begin = int(ids[0])
      id_end = id_begin
    elif len(ids) == 2:
      id_begin = int(ids[0])
      id_end = int(ids[1])
    else:
      raise ValueError(f"Invalid id format {ids}")

    yield from range(id_begin, id_end + 1)


from pathlib import Path
from typing import TypedDict

from ...pdf import PDFPageExtractor
from ..reporter import Reporter, AnalysingStep
from ..utils import Context
from .extractor import extract_ocr_page_xmls


class _State(TypedDict):
  completed_scanning: bool
  completed_pages: list[int]

def generate_ocr_pages(
      extractor: PDFPageExtractor,
      reporter: Reporter,
      pdf_path: Path,
      ocr_path: Path,
      assets_path: Path,
    ) -> None:

  context: Context[_State] = Context(
    reporter=reporter,
    path=ocr_path,
    init=lambda: {
      "completed_scanning": False,
      "completed_pages": [],
    },
  )
  if context.state["completed_scanning"]:
    return

  for path in (context.path, assets_path):
    path.mkdir(parents=True, exist_ok=True)

  reporter.go_to_step(AnalysingStep.OCR)

  for page_index, page_xml in extract_ocr_page_xmls(
    extractor=extractor,
    reporter=reporter,
    pdf_path=pdf_path,
    expected_page_indexes=set(context.state["completed_pages"]),
    cover_path=assets_path / "cover.png",
    assets_dir_path=assets_path,
  ):
    file_name = f"page_{page_index + 1}.xml"
    file_path = context.path / file_name
    context.write_xml_file(file_path, page_xml)
    context.state = {
      **context.state,
      "completed_pages": sorted([
        *context.state["completed_pages"],
        page_index,
      ]),
    }
  context.state = {
    **context.state,
    "completed_pages": [],
    "completed_scanning": True,
  }


import fitz

from pathlib import Path
from typing import Generator, Iterable
from PIL.Image import Image
from xml.etree.ElementTree import fromstring, Element, ParseError

from ...utils import sha256_hash
from .asset_matcher import search_asset_tags, AssetMatcher, AssetKind
from ..reporter import Reporter
from ...pdf import (
  PDFPageExtractor,
  Block,
  Text,
  TextKind,
  TextBlock,
  TableBlock,
  TableFormat,
  FigureBlock,
  FormulaBlock,
)


def extract_ocr_page_xmls(
    extractor: PDFPageExtractor,
    pdf_path: Path,
    expected_page_indexes: set[int],
    cover_path: Path,
    assets_dir_path: Path,
    reporter: Reporter,
  ) -> Generator[tuple[int, Element], None, None]:

  with fitz.open(pdf_path) as pdf:
    reporter.set(max_count=pdf.page_count)
    reporter.set_progress(len(expected_page_indexes))

    for i, blocks, image in extractor.extract_enumerated_blocks_and_image(
      pdf=pdf,
      page_indexes=(i for i in range(pdf.page_count) if i not in expected_page_indexes),
    ):
      if image and i == 0:
        image.save(cover_path)

      page_xml = _transform_page_xml(blocks)
      _migrate_expressions_and_save_images(
        root=page_xml,
        blocks=blocks,
        assets_dir_path=assets_dir_path,
      )
      reporter.increment()
      yield i, page_xml

def _transform_page_xml(blocks: list[Block]) -> Element:
  root = Element("page")
  for block in blocks:
    if isinstance(block, TextBlock):
      tag_name: str
      if block.kind == TextKind.TITLE:
        tag_name = "headline"
      elif block.kind == TextKind.PLAIN_TEXT:
        tag_name = "text"
      elif block.kind == TextKind.ABANDON:
        tag_name = "abandon"
      else:
        raise ValueError(f"Unknown text kind: {block.kind}")

      text_dom = Element(tag_name)
      if block.kind == TextKind.PLAIN_TEXT:
        text_dom.set("indent", "true" if block.has_paragraph_indentation else "false")
        text_dom.set("touch-end", "true" if block.last_line_touch_end else "false")

      _extends_line_doms(text_dom, block.texts)
      root.append(text_dom)

    elif isinstance(block, TableBlock):
      _append_asset_dom(root, block, "table")

    elif isinstance(block, FormulaBlock):
      _append_asset_dom(root, block, "formula")

    elif isinstance(block, FigureBlock):
      _append_asset_dom(root, block, "figure")

  return root

def _append_asset_dom(root: Element, block: Block, tag_name: str):
  root.append(Element(tag_name))
  if len(block.texts) > 0:
    caption_dom = Element(f"{tag_name}-caption")
    _extends_line_doms(caption_dom, block.texts)
    root.append(caption_dom)

def _extends_line_doms(parent: Element, texts: list[Text]):
  for text in texts:
    line_dom = Element("line")
    line_dom.set("confidence", "{:.2f}".format(text.rank))
    line_dom.text = text.content.replace("\n", " ")
    parent.append(line_dom)

def _migrate_expressions_and_save_images(root: Element, blocks: list[Block], assets_dir_path: Path):
  asset_matcher = AssetMatcher()
  images: dict[str, Image] = {}

  def register_image_and_get_hash(image: Image):
    hash = sha256_hash(image.tobytes())
    images[hash] = image
    return hash

  def create_children(tag_name: str, text: str) -> Iterable[Element]:
    child = Element(tag_name)
    child.text = text
    return (child,)

  for block in blocks:
    kind: AssetKind | None = None
    hash: str | None = None
    children: Iterable[Element] | None = None

    if isinstance(block, TableBlock):
      kind = AssetKind.TABLE
      hash = register_image_and_get_hash(block.image)
      if block.format == TableFormat.LATEX:
        children = create_children("latex", block.content)
      elif block.format == TableFormat.MARKDOWN:
        children = create_children("markdown", block.content)
      elif block.format == TableFormat.HTML:
        try:
          table_element = fromstring(block.content)
          table_content = Element("html")
          table_content.append(table_element)
          children = (table_content,)
        except ParseError as err:
          print("warn: parsed table HTML failed: ", err)
          pass

    elif isinstance(block, FormulaBlock):
      kind = AssetKind.FORMULA
      hash = register_image_and_get_hash(block.image)
      if block.content is not None:
        children = create_children("latex", block.content)

    elif isinstance(block, FigureBlock):
      kind = AssetKind.FIGURE
      hash = register_image_and_get_hash(block.image)

    if kind is not None:
      asset_matcher.register_virtual_dom(
        kind=kind,
        hash=hash,
        children=children,
      )

  asset_matcher.recover_asset_doms_for_xml(root)

  for asset_dom in search_asset_tags(root):
    hash = asset_dom.get("hash", None)
    if hash is None:
      continue
    image: Image | None = images.get(hash, None)
    if image is None:
      continue
    file_path = assets_dir_path / f"{hash}.png"
    if file_path.exists():
      continue
    image.save(file_path, "PNG")


from typing import Callable
from enum import auto, Enum
from threading import Lock


class AnalysingStep(Enum):
  OCR = auto()
  EXTRACT_SEQUENCE = auto()
  VERIFY_TEXT_PARAGRAPH = auto()
  VERIFY_FOOTNOTE_PARAGRAPH = auto()
  CORRECT_TEXT = auto()
  CORRECT_FOOTNOTE = auto()
  EXTRACT_META = auto()
  COLLECT_CONTENTS = auto()
  ANALYSE_CONTENTS = auto()
  MAPPING_CONTENTS = auto()
  GENERATE_FOOTNOTES = auto()
  OUTPUT = auto()

# func(completed_count: int, max_count: int | None) -> None
AnalysingProgressReport = Callable[[int, int | None], None]

# func(step: AnalysingStep) -> None
AnalysingStepReport = Callable[[AnalysingStep], None]

# thread safe
class Reporter:
  def __init__(
      self,
      report_step: AnalysingStepReport | None,
      report_progress: AnalysingProgressReport | None,
    ) -> None:

    self._lock: Lock = Lock()
    self._report_step: AnalysingStepReport | None = report_step
    self._report_progress: AnalysingProgressReport | None = report_progress
    self._progress: int = 0
    self._max_progress_count: int | None = None

  def go_to_step(self, step: AnalysingStep) -> None:
    with self._lock:
      if self._report_step is not None:
        self._report_step(step)
      self._progress = 0
      self._max_progress_count = None

  def set(self, max_count: int) -> None:
    with self._lock:
      self._max_progress_count = max_count

  def set_progress(self, progress: int) -> None:
    with self._lock:
      if self._max_progress_count is not None:
        progress = min(progress, self._max_progress_count)

      if progress == self._progress:
        return
      self._progress = progress
      if self._report_progress is None:
        return

      self._report_progress(self._progress, self._max_progress_count)

  def increment(self, count: int = 1) -> None:
    self.set_progress(self._progress + count)




from __future__ import annotations
from threading import Lock
from typing import Any, TypeVar, Generic, Sequence, Iterator, Callable, Generator
from .context import Context


T = TypeVar("T", bound=tuple[int, ...])
S = TypeVar("S", bound=dict[str, Any])
P = TypeVar("P")

_STATE_KEY = "completed_ranges"

# thread safe
class Partition(Generic[T, S, P]):
  def __init__(
        self,
        dimension: int,
        context: Context[S],
        sequence: Sequence[tuple[T, T, P]],
        remove: Callable[[T, T], None] | None = None,
        done: Callable[[T, T], None] | None = None,
      ) -> None:

    super().__init__()
    self._dimension: int = dimension
    self._context: Context[S] = context
    self._range_iterator: Iterator[tuple[T | int, T | int, P]] = iter(sequence)
    self._done: Callable[[T, T], None] = done or (lambda _, __: None)
    self._remove: Callable[[T, T], None] = remove or (lambda _, __: None)
    self._iter_lock: Lock = Lock()
    self._range_lock: Lock = Lock()
    self._last_index: T | None = None
    self._done_ranges: list[tuple[T, T]] = []
    self._to_remove_ranges: list[tuple[T, T]] = []

    for item in context.state.get(_STATE_KEY, ()):
      half_len = len(item) // 2
      begin: list[int] = []
      end: list[int] = []
      for i in range(half_len):
        begin.append(item[i])
        end.append(item[i + half_len])
      self._done_ranges.append((
        tuple(begin),
        tuple(end),
      ))

    self._done_ranges.sort(key=lambda x: x[0])

  def __enter__(self) -> Partition[T, S, P]:
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    with self._range_lock:
      self._done_ranges.sort(key=lambda x: x[0])
      if exc_type is None and self._last_index is not None:
        last_index: int = -1
        for i, (begin, end) in enumerate(self._done_ranges):
          if self._last_index < max(begin, end):
            last_index = i
            break
        if last_index >= 0:
          self._done_ranges = self._done_ranges[:last_index]
          self._sync_done_ranges()

      for begin, end in self._to_remove_ranges:
        self._remove(begin, end)
      self._to_remove_ranges.clear()

    return False

  def pop_tasks(self) -> Generator[PartitionTask[T, S, P], None, None]:
    while True:
      task = self.pop_task()
      if task is None:
        break
      yield task

  def pop_task(self) -> PartitionTask[T, S, P] | None:
    with self._iter_lock:
      while True:
        range = next(self._range_iterator, None)
        if range is None:
          return None

        begin, end, payload = range
        begin = self._assert_index(begin)
        end = self._assert_index(end)

        if (begin, end) in self._done_ranges:
          self._done(begin, end)
          continue

        with self._range_lock:
          if self._last_index is None:
            self._last_index = max(begin, end)
          else:
            self._last_index = max(begin, end, self._last_index)

        return PartitionTask(
          begin=begin,
          end=end,
          payload=payload,
          done=lambda: self._on_task_done(begin, end),
        )

  def _assert_index(self, index: T | int) -> T:
    if isinstance(index, int):
      index = (index,)
    else:
      assert isinstance(index, tuple), f"Index must be a tuple, got {type(index)}"

    assert len(index) == self._dimension, f"Index must have length {self._dimension}, got {len(index)}"
    for i, x in enumerate(index):
      assert isinstance(x, int), f"Index[{i}] must be an int, got {type(x)}"

    return index

  def _on_task_done(self, done_begin: T, done_end: T) -> None:
    with self._range_lock:
      found_matched = False
      new_done_ranges: list[tuple[T, T]] = []
      for begin, end in self._done_ranges:
        if done_begin <= end and done_end >= begin:
          self._to_remove_ranges.append((begin, end))
        else:
          if done_begin == begin and done_end == end:
            found_matched = True
          new_done_ranges.append((begin, end))

      if not found_matched:
        new_done_ranges.append((done_begin, done_end))

      new_done_ranges.sort(key=lambda x: x[0])
      self._done_ranges = new_done_ranges
      self._sync_done_ranges()
      self._done(done_begin, done_end)

  def _sync_done_ranges(self) -> None:
    ranges: list[list[int]] = []
    self._done_ranges.sort(key=lambda x: x[0])
    for begin, end in self._done_ranges:
      ranges.append([*begin, *end])
    self._context.state = {
      **self._context.state,
      _STATE_KEY: ranges,
    }

class PartitionTask(Generic[T, S, P]):
  def __init__(self, begin: T, end: T, payload: P, done: Callable[[], None]) -> None:
    super().__init__()
    self.begin: T = begin
    self.end: T = end
    self.payload: P = payload
    self._done: Callable[[], None] = done

  def __enter__(self) -> PartitionTask[T, S, P]:
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    if exc_type is None:
      self._done()
    return False



import re

from pathlib import Path
from strenum import StrEnum
from xml.etree.ElementTree import Element

from ...llm import LLM
from ...xml import clone as clone_xml
from ..reference import samples, NumberStyle
from ..utils import read_xml_file, Context
from .common import State


def repeat_multiple_correct(
      llm: LLM,
      context: Context[State],
      save_path: Path,
      raw_request: Element,
      is_footnote: bool,
    ) -> Element:

  save_path.mkdir(parents=True, exist_ok=True)
  repeater = _Repeater(
    llm=llm,
    context=context,
    save_path=save_path,
    is_footnote=is_footnote,
  )
  return repeater.do(raw_request)

_BASIC_RETRY_TIMES = 2
_FILE_NAME_PATTERN = re.compile(r"^step_(\d+)\.xml$")

class _Quality(StrEnum):
  PERFECT = "perfect"
  GOOD = "good"
  FAIR = "fair"
  POOR = "poor"
  INVALID = "invalid"

_STEPS_INCREMENT: tuple[tuple[_Quality, int], ...] = (
  (_Quality.INVALID, 5),
  (_Quality.POOR, 4),
  (_Quality.FAIR, 3),
  (_Quality.GOOD, 0),
  (_Quality.PERFECT, 0),
)

class _Repeater:
  def __init__(
        self,
        llm: LLM,
        context: Context[State],
        save_path: Path,
        is_footnote: bool,
      ):
    self._llm: LLM = llm
    self._ctx: Context[State] = context
    self._save_path: Path = save_path
    self._is_footnote: bool = is_footnote
    self._quality: _Quality | None = None
    self._remain_steps: int = _BASIC_RETRY_TIMES
    self._next_index: int = 1

  def do(self, raw_request_element: Element) -> Element:
    request_element = self._recover_state(raw_request_element)
    while self._remain_steps > 0 and \
          self._quality != _Quality.PERFECT:

      resp_element = self._llm.request_xml(
        template_name="correction/multiple",
        user_data=request_element,
        params={
          "layouts_count": 4,
          "is_footnote": self._is_footnote,
          "marks": samples(NumberStyle.CIRCLED_NUMBER, 6),
        },
      )
      quality = self._read_quality_from_element(resp_element)
      self._report_quality(quality)
      updation_element = self._save_step_file(
        request_element=request_element,
        resp_element=resp_element,
      )
      request_element = self._update_request_element(
        raw_request_element=request_element,
        updation_element=updation_element,
      )

    return request_element

  def _recover_state(self, raw_request_element: Element) -> Element:
    request_element: Element = raw_request_element
    files: list[tuple[int, Path]] = []

    for file in self._save_path.iterdir():
      match = _FILE_NAME_PATTERN.match(file.name)
      if match:
        index = int(match.group(1))
        files.append((index, file))

    for index, file in sorted(files, key=lambda x: x[0]):
      file_element = read_xml_file(file)
      request_element = file_element.find("request") or request_element
      file_quality = self._read_quality_from_element(file_element)
      self._report_quality(file_quality)
      self._next_index = index + 1

    return request_element

  def _save_step_file(self, request_element: Element, resp_element: Element) -> Element:
    file_element = Element("correction")
    overview_element = resp_element.find("overview")
    if overview_element is not None:
      resp_element.remove(overview_element)
      file_element.append(overview_element)

    if len(resp_element) > 0:
      resp_element.tag = "updation"
      file_element.append(resp_element)
      file_element.append(request_element)

    file_name = f"step_{self._next_index}.xml"
    file_path = self._save_path / file_name
    self._ctx.write_xml_file(file_path, file_element)
    self._next_index += 1

    return resp_element

  def _read_quality_from_element(self, element: Element) -> _Quality:
    quality: _Quality = _Quality.PERFECT
    overview_element = element.find("overview")
    if overview_element is not None:
      quality = _Quality(overview_element.get("quality", _Quality.PERFECT.value))
    return quality

  def _report_quality(self, resp_quality: _Quality) -> None:
    if self._quality is None:
      self._quality = resp_quality

    else:
      found_begin = False
      found_end = False
      increment_remain_steps: int = 0

      for quality, increment in _STEPS_INCREMENT:
        if quality == self._quality:
          found_begin = True
        if quality == resp_quality:
          found_end = True
          break
        if found_begin:
          increment_remain_steps += increment

      if found_begin and found_end:
        self._remain_steps += increment_remain_steps
        self._quality = resp_quality

    self._remain_steps = max(0, self._remain_steps - 1)

  def _update_request_element(self, raw_request_element: Element, updation_element: Element) -> Element:
    # TODO: 此处可用传统逻辑设计些防御性操作，以避免 LLM 生成内容发生如下错误
    #       这些错误我虽然用 prompt 进行了约束，但无法杜绝：
    #         1. layout ID 错误，这将导致整个自然段丢失
    #         2. line 丢失某些行，或插入某些行
    #
    #       可以考虑用字符串匹配或 Sentence embedding 技术来发现这些问题，并用传统算法修复

    request_ids: list[str] = []
    request_layouts: dict[str, Element] = {}

    for layout_element in clone_xml(raw_request_element):
      layout_id = layout_element.get("id", None)
      if layout_id is not None:
        request_ids.append(layout_id)
        request_layouts[layout_id] = layout_element

    for updation_layout_element in updation_element:
      layout_id = updation_layout_element.get("id", None)
      if layout_id is None:
        continue
      request_layout_element = request_layouts.get(layout_id, None)
      if request_layout_element is None:
        continue
      raw_attrib = request_layout_element.attrib
      request_layout_element.clear()
      request_layout_element.extend(updation_layout_element)
      request_layout_element.attrib = raw_attrib

    request_element = Element("request")
    for request_id in request_ids:
      layout_element = request_layouts.get(request_id, None)
      if layout_element is not None:
        request_element.append(layout_element)

    return request_element



from typing import cast, Any, Callable
from io import StringIO
from time import sleep
from pydantic import SecretStr
from logging import Logger
from langchain_core.messages import SystemMessage, HumanMessage, AIMessage
from langchain_core.language_models import LanguageModelInput
from langchain_openai import ChatOpenAI

from .increasable import Increasable, Increaser
from .error import is_retry_error


class LLMExecutor:
  def __init__(
    self,
    api_key: SecretStr,
    url: str,
    model: str,
    timeout: float | None,
    top_p: Increasable,
    temperature: Increasable,
    retry_times: int,
    retry_interval_seconds: float,
    create_logger: Callable[[], Logger | None],
  ) -> None:

    self._timeout: float | None = timeout
    self._top_p: Increasable = top_p
    self._temperature: Increasable = temperature
    self._retry_times: int = retry_times
    self._retry_interval_seconds: float = retry_interval_seconds
    self._create_logger: Callable[[], Logger | None] = create_logger
    self._model = ChatOpenAI(
      api_key=cast(SecretStr, api_key),
      base_url=url,
      model=model,
      timeout=timeout,
    )

  def request(self, input: LanguageModelInput, parser: Callable[[str], Any]) -> Any:
    result: Any | None = None
    last_error: Exception | None = None
    did_success = False
    top_p: Increaser = self._top_p.context()
    temperature: Increaser = self._temperature.context()
    logger = self._create_logger()

    if logger is not None:
      logger.debug(f"[[Request]]:\n{self._input2str(input)}\n")

    try:
      for i in range(self._retry_times + 1):
        try:
          response = self._invoke_model(
            input=input,
            top_p=top_p.current,
            temperature=temperature.current,
          )
          if logger is not None:
            logger.debug(f"[[Response]]:\n{response}\n")

        except Exception as err:
          last_error = err
          if not is_retry_error(err):
            raise err
          if logger is not None:
            logger.warning(f"request failed with connection error, retrying... ({i + 1} times)")
          if self._retry_interval_seconds > 0.0 and \
            i < self._retry_times:
            sleep(self._retry_interval_seconds)
          continue

        try:
          result = parser(response)
          did_success = True
          break

        except Exception as err:
          last_error = err
          warn_message = f"request failed with parsing error, retrying... ({i + 1} times)"
          if logger is not None:
            logger.warning(warn_message)
          print(warn_message)
          top_p.increase()
          temperature.increase()
          if self._retry_interval_seconds > 0.0 and \
            i < self._retry_times:
            sleep(self._retry_interval_seconds)
          continue

    except KeyboardInterrupt as err:
      if last_error is not None and logger is not None:
        logger.debug(f"[[Error]]:\n{last_error}\n")
      raise err

    if not did_success:
      if last_error is None:
        raise RuntimeError("Request failed with unknown error")
      else:
        raise last_error

    return result

  def _input2str(self, input: LanguageModelInput) -> str:
    if isinstance(input, str):
      return input
    if not isinstance(input, list):
      raise ValueError(f"Unsupported input type: {type(input)}")

    buffer = StringIO()
    is_first = True
    for message in input:
      if not is_first:
        buffer.write("\n\n")
      if isinstance(message, SystemMessage):
        buffer.write("System:\n")
        buffer.write(message.content)
      elif isinstance(message, HumanMessage):
        buffer.write("User:\n")
        buffer.write(message.content)
      elif isinstance(message, AIMessage):
        buffer.write("Assistant:\n")
        buffer.write(message.content)
      else:
        buffer.write(str(message))
      is_first = False

    return buffer.getvalue()

  def _invoke_model(
        self,
        input: LanguageModelInput,
        top_p: float | None,
        temperature: float | None,
      ):
    stream = self._model.stream(
      input=input,
      timeout=self._timeout,
      top_p=top_p,
      temperature=temperature,
    )
    buffer = StringIO()
    for chunk in stream:
      data = str(chunk.content)
      buffer.write(data)
    return buffer.getvalue()


from xml.etree.ElementTree import Element

def clone(element: Element) -> Element:
  new_element = Element(element.tag)
  for attr_name, attr_value in element.items():
    new_element.set(attr_name, attr_value)
  new_element.text = element.text
  for child in element:
    new_child = clone(child)
    new_element.append(new_child)
    new_child.tail = child.tail
  return new_element


from io import StringIO
from typing import Generator, Iterable
from enum import auto, Enum
from .tag import is_valid_name_char, is_valid_value_char, Tag, TagKind


_SPACES = (" ", "\n")

class _Phase(Enum):
  OUTSIDE = auto()
  LEFT_BRACKET = auto()
  LEFT_SLASH = auto()
  TAG_NAME = auto()
  TAG_GAP = auto()
  ATTRIBUTE_NAME = auto()
  ATTRIBUTE_NAME_EQUAL = auto()
  ATTRIBUTE_VALUE = auto()
  MUST_CLOSING_SIGN = auto()

class _ParsedResult(Enum):
  Continue = auto()
  Success = auto()
  Failed = auto()

def parse_tags(chars: Iterable[str]) -> Generator[str | Tag, None, None]:
  yield from _XMLTagsParser().do(chars)

class _XMLTagsParser:
  def __init__(self):
    self._outside_buffer: StringIO = StringIO()
    self._tag_buffer: StringIO = StringIO()
    self._tag: Tag | None = None
    self._phase: _Phase = _Phase.OUTSIDE

  def do(self, chars: Iterable[str]) -> Generator[str | Tag, None, None]:
    for char in chars:
      parsed_result = self._parse_char(char)
      yield from self._generate_by_result(parsed_result)

    self._outside_buffer.write(self._tag_buffer.getvalue())
    outside_text = self._outside_buffer.getvalue()
    if outside_text != "":
      yield outside_text

  def _parse_char(self, char: str) -> _ParsedResult:
    parsed_result: _ParsedResult = _ParsedResult.Continue

    if self._phase == _Phase.OUTSIDE:
      if char != "<":
        self._outside_buffer.write(char)
      else:
        self._phase = _Phase.LEFT_BRACKET
        self._tag_buffer.write(char)
        self._tag = Tag(
          kind=TagKind.OPENING,
          name="",
          proto="",
          attributes=[],
        )
    else:
      self._tag_buffer.write(char)

      if self._phase == _Phase.LEFT_BRACKET:
        if char == "/":
          self._tag.kind = TagKind.CLOSING
          self._phase = _Phase.LEFT_SLASH
        elif is_valid_name_char(char):
          self._tag.name += char
          self._phase = _Phase.TAG_NAME
        else:
          parsed_result = _ParsedResult.Failed

      elif self._phase == _Phase.LEFT_SLASH:
        if is_valid_name_char(char):
          self._tag.name += char
          self._phase = _Phase.TAG_NAME
        else:
          parsed_result = _ParsedResult.Failed

      elif self._phase == _Phase.TAG_NAME:
        if char in _SPACES:
          self._phase = _Phase.TAG_GAP
        elif is_valid_name_char(char):
          self._tag.name += char
        elif char == ">":
          parsed_result = _ParsedResult.Success
        elif char == "/" and self._tag.kind == TagKind.OPENING:
          self._tag.kind = TagKind.SELF_CLOSING
          self._phase = _Phase.MUST_CLOSING_SIGN
        else:
          parsed_result = _ParsedResult.Failed

      elif self._phase == _Phase.TAG_GAP:
        if char in _SPACES:
          pass
        elif is_valid_name_char(char):
          self._tag.attributes.append((char, ""))
          self._phase = _Phase.ATTRIBUTE_NAME
        elif char == ">":
          parsed_result = _ParsedResult.Success
        elif char == "/" and self._tag.kind == TagKind.OPENING:
          self._tag.kind = TagKind.SELF_CLOSING
          self._phase = _Phase.MUST_CLOSING_SIGN
        else:
          parsed_result = _ParsedResult.Failed

      elif self._phase == _Phase.ATTRIBUTE_NAME:
        if is_valid_name_char(char):
          attr_name, attr_value = self._tag.attributes[-1]
          attr_name = attr_name + char
          self._tag.attributes[-1] = (attr_name, attr_value)
        elif char == "=":
          self._phase = _Phase.ATTRIBUTE_NAME_EQUAL
        else:
          parsed_result = _ParsedResult.Failed

      elif self._phase == _Phase.ATTRIBUTE_NAME_EQUAL:
        if char == "\"":
          self._phase = _Phase.ATTRIBUTE_VALUE
        else:
          parsed_result = _ParsedResult.Failed

      elif self._phase == _Phase.ATTRIBUTE_VALUE:
        if is_valid_value_char(char):
          attr_name, attr_value = self._tag.attributes[-1]
          attr_value = attr_value + char
          self._tag.attributes[-1] = (attr_name, attr_value)
        elif char == "\"":
          self._phase = _Phase.TAG_GAP
        else:
          parsed_result = _ParsedResult.Failed

      elif self._phase == _Phase.MUST_CLOSING_SIGN:
        if char == ">":
          parsed_result = _ParsedResult.Success
        else:
          parsed_result = _ParsedResult.Failed

    return parsed_result

  def _generate_by_result(self, parsed_result: _ParsedResult) -> Generator[str | Tag, None, None]:
    if parsed_result == _ParsedResult.Success:
      assert self._tag is not None
      if self._is_tag_valid(self._tag):
        outside_text = self._outside_buffer.getvalue()
        self._clear_buffer(self._outside_buffer)
        self._clear_buffer(self._tag_buffer)
        if outside_text != "":
          yield outside_text
        yield self._tag
      else:
        self._tag.proto = self._tag_buffer.getvalue()
        self._outside_buffer.write(self._tag.proto)
        self._clear_buffer(self._tag_buffer)
      self._tag = None
      self._phase = _Phase.OUTSIDE

    elif parsed_result == _ParsedResult.Failed:
      self._outside_buffer.write(self._tag_buffer.getvalue())
      self._clear_buffer(self._tag_buffer)
      self._phase = _Phase.OUTSIDE

  def _is_tag_valid(self, tag: Tag) -> bool:
    if tag.kind == TagKind.CLOSING and len(tag.attributes) > 0:
      return False
    if tag.find_invalid_name() is not None:
      return False
    return True

  def _clear_buffer(self, buffer: StringIO):
    buffer.truncate(0)
    buffer.seek(0)


from dataclasses import dataclass

@dataclass
class LLMWindowTokens:
  max_request_data_tokens: int = 4096
  max_verify_paragraph_tokens: int = 512,
  max_verify_paragraphs_count: int = 8,

_MIN_TOKENS = 20

def parse_window_tokens(window_tokens: LLMWindowTokens | int | None) -> LLMWindowTokens:
  if window_tokens is None:
    window_tokens = 4096

  if isinstance(window_tokens, int):
    if window_tokens <= _MIN_TOKENS:
      raise ValueError(f"window_tokens must be greater than {_MIN_TOKENS}.")

    max_verify_paragraph_tokens = max(_MIN_TOKENS, window_tokens // 8)
    max_verify_paragraphs_count = max(2, window_tokens // max_verify_paragraph_tokens)
    window_tokens = LLMWindowTokens(
      max_request_data_tokens=window_tokens,
      max_verify_paragraph_tokens=max_verify_paragraph_tokens,
      max_verify_paragraphs_count=max_verify_paragraphs_count,
    )

  if isinstance(window_tokens, LLMWindowTokens):
    if window_tokens.max_request_data_tokens <= _MIN_TOKENS:
      raise ValueError(f"max_request_data_tokens must be greater than {_MIN_TOKENS}.")
    if window_tokens.max_verify_paragraph_tokens <= _MIN_TOKENS:
      raise ValueError(f"max_verify_paragraph_tokens must be greater than {_MIN_TOKENS}.")
    if window_tokens.max_verify_paragraphs_count < 1:
      raise ValueError("max_verify_paragraphs_count must be at least 1.")

  return window_tokens


from pathlib import Path
from json import dumps as json_dumps
from xml.etree.ElementTree import Element

from ...llm import LLM
from ...xml import encode_friendly
from ..sequence import read_paragraphs


def extract_meta(
      llm: LLM,
      workspace_path: Path,
      sequence_path: Path,
      max_request_tokens: int = 4090,
    ) -> Path:

  meta_path = workspace_path / "meta.json"

  if not meta_path.exists():
    workspace_path.mkdir(parents=True, exist_ok=True)
    request_element = _create_request_element(
      llm=llm,
      sequence_path=sequence_path,
      max_request_tokens=max_request_tokens,
    )
    meta_json = llm.request_json(
      template_name="meta",
      user_data=request_element,
    )
    with open(meta_path, "w", encoding="utf-8") as file:
      file.write(json_dumps(
        meta_json,
        ensure_ascii=False,
        indent=2,
      ))

  return meta_path

def _create_request_element(
      llm: LLM,
      sequence_path: Path,
      max_request_tokens: int = 4090,
    ) -> Element:

  tokens: int = 0
  request_element = Element("request")
  page_element: Element | None = None
  page_index: int = -1

  for paragraph, layout_element in _extract_layout_elements(sequence_path):
    layout_tokens = llm.count_tokens_count(encode_friendly(layout_element))
    if tokens > 0 and tokens + layout_tokens > max_request_tokens:
      break
    if page_index != paragraph.page_index:
      if page_element:
        request_element.append(page_element)
      page_element = Element("page")
      page_element.set("page-index", str(paragraph.page_index))
      page_element.set("type", paragraph.type.value)
      page_index = paragraph.page_index

    page_element.append(layout_element)
    tokens += layout_tokens

  if page_element:
    request_element.append(page_element)
  return request_element

def _extract_layout_elements(sequence_path: Path):
  for paragraph in read_paragraphs(sequence_path):
    layout_element: Element | None = None
    line_texts: list[str] = []

    for layout in paragraph.layouts:
      if layout_element is None:
        layout_element = Element(layout.kind.value)
      for line in layout.lines:
        line_texts.append(line.text)

    if layout_element is not None:
      layout_element.text = "\n".join(line_texts)
      yield paragraph, layout_element


import openai
import httpx
import requests


def is_retry_error(err: Exception) -> bool:
  if _is_openai_retry_error(err):
    return True
  if _is_httpx_retry_error(err):
    return True
  if _is_request_retry_error(err):
    return True
  return False

# https://help.openai.com/en/articles/6897213-openai-library-error-types-guidance
def _is_openai_retry_error(err: Exception) -> bool:
  if isinstance(err, openai.Timeout):
    return True
  if isinstance(err, openai.APIConnectionError):
    return True
  if isinstance(err, openai.InternalServerError):
    return err.status_code in (502, 503, 504)
  return False

# https://www.python-httpx.org/exceptions/
def _is_httpx_retry_error(err: Exception) -> bool:
  if isinstance(err, httpx.RemoteProtocolError):
    return True
  if isinstance(err, httpx.StreamError):
    return True
  if isinstance(err, httpx.TimeoutException):
    return True
  if isinstance(err, httpx.NetworkError):
    return True
  if isinstance(err, httpx.ProtocolError):
    return True
  return False

# https://requests.readthedocs.io/en/latest/api/#exceptions
def _is_request_retry_error(err: Exception) -> bool:
  if isinstance(err, requests.ConnectionError):
    return True
  if isinstance(err, requests.ConnectTimeout):
    return True
  if isinstance(err, requests.ReadTimeout):
    return True
  if isinstance(err, requests.Timeout):
    return True
  return False


from dataclasses import dataclass
from pathlib import Path
from typing import Generator
from xml.etree.ElementTree import fromstring, Element

from .common import get_truncation_attr, State, SequenceType, Truncation
from .draft import TruncationKind, ParagraphDraft
from ...llm import LLM
from ...xml import encode, encode_friendly
from ..data import ParagraphType
from ..utils import (
  remove_file,
  read_xml_file,
  xml_files,
  Context,
  Partition,
  PartitionTask,
  MultiThreads,
)


def join(
      llm: LLM,
      context: Context[State],
      threads: MultiThreads,
      type: SequenceType,
      extraction_path: Path,
      join_path: Path,
    ) -> None:

  joint = _Joint(
    llm=llm,
    context=context,
    threads=threads,
    type=type,
    extraction_path=extraction_path,
    join_path=join_path,
  )
  joint.do()

@dataclass
class _SequenceMeta:
  paragraph_type: ParagraphType
  page_index: int
  truncations: tuple[Truncation, Truncation]

_MetaTruncationDict = dict[int, tuple[_SequenceMeta, TruncationKind]]

class _Joint:
  def __init__(
        self,
        llm: LLM,
        context: Context[State],
        threads: MultiThreads,
        type: SequenceType,
        extraction_path: Path,
        join_path: Path,
      ) -> None:

    self._llm: LLM = llm
    self._ctx: Context[State] = context
    self._threads: MultiThreads = threads
    self._type: SequenceType = type
    self._extraction_path: Path = extraction_path
    self._join_path: Path = join_path

  def do(self):
    metas = self._extract_sequence_metas()
    truncations = list(self._extract_truncations(metas))
    meta_truncation_dict: _MetaTruncationDict = {}

    for i, meta in enumerate(metas):
      truncation: TruncationKind = TruncationKind.NO
      if i < len(truncations):
        truncation = truncations[i]
      meta_truncation_dict[meta.page_index] = (meta, truncation)

    self._request_llm_to_verify(meta_truncation_dict)

    last_page_index = 0
    next_paragraph_id = 1

    for paragraph in self._generate_paragraphs(meta_truncation_dict):
      page_index = paragraph.page_index
      if last_page_index != page_index:
        last_page_index = page_index
        next_paragraph_id = 1

      save_dir_path = self._ctx.path.joinpath("output", self._type.value)
      save_dir_path.mkdir(parents=True, exist_ok=True)

      paragraph_id = f"{page_index}_{next_paragraph_id}"
      file_path = save_dir_path / f"paragraph_{paragraph_id}.xml"
      next_paragraph_id += 1

      self._ctx.atomic_write(
        file_path=file_path,
        content=encode(paragraph.to_xml()),
      )

  def _extract_sequence_metas(self) -> list[_SequenceMeta]:
    metas: list[_SequenceMeta] = []
    for page_index, sequence in self._extract_sequences():
      truncation_begin = get_truncation_attr(sequence, "truncation-begin")
      truncation_end = get_truncation_attr(sequence, "truncation-end")
      metas.append(_SequenceMeta(
        paragraph_type=ParagraphType(sequence.get("type")),
        page_index=page_index,
        truncations=(truncation_begin, truncation_end),
      ))

    pre_page_index = 0 # page-index is begin from 1
    pre_meta: _SequenceMeta | None = None
    for meta in metas:
      if pre_page_index + 1 != meta.page_index:
        # The pages are not continuous, and it is impossible to cross pages in the middle,
        # so this assertion is made
        meta.truncations = (Truncation.NO, meta.truncations[1])
        if pre_meta is not None:
          pre_meta.truncations = (pre_meta.truncations[0], Truncation.NO)
      pre_page_index = meta.page_index
      pre_meta = meta

    return metas

  def _extract_truncations(self, metas: list[_SequenceMeta]):
    for i in range(0, len(metas) - 1):
      meta1 = metas[i]
      meta2 = metas[i + 1]
      _, truncation1 = meta1.truncations
      truncation2, _ = meta2.truncations
      truncations = (truncation1, truncation2)

      if all(t == Truncation.NO for t in truncations):
        yield TruncationKind.NO
        continue

      if any(t == Truncation.YES for t in truncations) and \
         all(t in (Truncation.YES, Truncation.PROBABLY) for t in truncations):
        yield TruncationKind.VERIFIED
        continue

      yield TruncationKind.UNCERTAIN

  def _request_llm_to_verify(self, meta_truncation_dict: _MetaTruncationDict):
    uncertain_truncations_count = sum(
      1 for _, kind in meta_truncation_dict.values()
      if kind == TruncationKind.UNCERTAIN
    )
    self._join_path.mkdir(parents=True, exist_ok=True)
    self._ctx.reporter.set(uncertain_truncations_count)
    partition: Partition[tuple[int], State, Element] = Partition(
      dimension=1,
      context=self._ctx,
      sequence=(
        (min(page_indexes), max(page_indexes), request_element)
        for page_indexes, request_element in self._search_uncertain_request(
          meta_truncation_dict=meta_truncation_dict,
        )
      ),
      done=lambda begin, end: self._ctx.reporter.increment(
        count=len(read_xml_file(
          file_path=self._join_path / f"truncation_{begin[0]}_{end[0]}.xml"
        ))
      ),
      remove=lambda begin, end: remove_file(
        file_path=self._join_path / f"truncation_{begin[0]}_{end[0]}.xml"
      ),
    )
    with partition:
      self._threads.run(
        next_task=partition.pop_task,
        thread_payload=lambda: None,
        invoke=lambda _, task: self._emit_request(task),
      )

    for file_path, file_prefix, _, _ in xml_files(self._extraction_path):
      if file_prefix != "truncation":
        continue

      for child in read_xml_file(file_path):
        if child.tag != "paragraph":
          continue
        page_index = int(child.get("first-page-index", "-1"))
        if page_index not in meta_truncation_dict:
          continue
        conclusion = child.get("conclusion", None)
        meta, _ = meta_truncation_dict[page_index]
        truncation = TruncationKind.NO
        if conclusion == Truncation.YES:
          truncation = TruncationKind.VERIFIED
        meta_truncation_dict[page_index] = (meta, truncation)

  def _search_uncertain_request(self, meta_truncation_dict: _MetaTruncationDict):
    max_verify_paragraphs_count = self._ctx.state["max_verify_paragraphs_count"]
    request_element: Element | None = None
    request_page_indexes: list[int] = []

    for page_index, text1, text2 in self._search_uncertain_texts(meta_truncation_dict):
      if request_element is not None and len(request_element) > max_verify_paragraphs_count:
        yield request_page_indexes, request_element
        request_element = None
        request_page_indexes = []

      if request_element is None:
        request_element = Element("request")

      paragraph_element = Element("paragraph")
      paragraph_element.set("first-page-index", str(page_index))
      paragraph_element.set("second-page-index", str(page_index + 1))

      for text in (text1, text2):
        text_element = Element("text")
        text_element.text = text
        paragraph_element.append(text_element)

      request_element.append(paragraph_element)
      request_page_indexes.append(page_index)

    if request_element is not None:
      yield request_page_indexes, request_element

  def _emit_request(self, task: PartitionTask[tuple[int], State, Element]) -> None:
    with task:
      begin = task.begin[0]
      end = task.end[0]
      request_element = task.payload
      resp_element = self._llm.request_xml(
        template_name="sequence/truncation",
        user_data=request_element,
        params={
          "count": len(request_element),
        },
      )
      self._ctx.write_xml_file(
        file_path=self._join_path / f"truncation_{begin}_{end}.xml",
        xml=resp_element,
      )

  def _search_uncertain_texts(self, meta_truncation_dict: _MetaTruncationDict):
    max_verify_paragraph_tokens = self._ctx.state["max_verify_paragraph_tokens"]
    tail: Element | None = None

    for page_index, sequence in self._extract_sequences():
      _, truncation = meta_truncation_dict[page_index]
      head, body = self._split_sequence(sequence)
      if tail is not None:
        text1 = "".join((line.text or "").strip() for line in tail)
        text2 = "".join((line.text or "").strip() for line in head)
        tail = None

        raw_tokens1 = self._llm.encode_tokens(text1)
        raw_tokens2 = list(reversed(self._llm.encode_tokens(text2)))
        tokens1: list[int] = []
        tokens2: list[int] = []

        while len(tokens1) + len(tokens2) < max_verify_paragraph_tokens:
          if raw_tokens1:
            tokens1.append(raw_tokens1.pop())
          if raw_tokens2:
            tokens2.append(raw_tokens2.pop())
          if not raw_tokens1 and not raw_tokens2:
            break

        text1 = self._llm.decode_tokens(list(reversed(tokens1)))
        text2 = self._llm.decode_tokens(tokens2)

        if raw_tokens1:
          text1 = "..." + text1
        if raw_tokens2:
          text2 = text2 + "..."

        yield page_index - 1, text1, text2

      if truncation == TruncationKind.UNCERTAIN:
        tail = head if len(body) == 0 else body[-1]

  def _generate_paragraphs(self, meta_truncation_dict: _MetaTruncationDict) -> Generator[ParagraphDraft, None, None]:
    max_request_data_tokens = self._ctx.state["max_request_data_tokens"]
    for paragraph in self._join_and_collect_paragraphs(meta_truncation_dict):
      if paragraph.tokens <= max_request_data_tokens:
        yield paragraph
      else:
        # TODO: 临时解决 Paragraph 过大导致后续请求拆分拆不动的问题
        #      https://github.com/oomol-lab/pdf-craft/issues/227
        for forked in paragraph.fork(max_request_data_tokens):
          if forked.tokens > max_request_data_tokens:
            print(f"Warning: paragraph at page {forked.page_index} has too many tokens: {forked.tokens}")
          yield forked

  def _join_and_collect_paragraphs(self, meta_truncation_dict: _MetaTruncationDict) -> Generator[ParagraphDraft, None, None]:
    last_paragraph: ParagraphDraft | None = None
    for page_index, sequence in self._extract_sequences():
      meta, truncation = meta_truncation_dict[page_index]
      head, body = self._split_sequence(sequence)

      if last_paragraph is not None and \
         meta.paragraph_type != last_paragraph.type:
        yield last_paragraph
        last_paragraph = None

      tokens = self._llm.count_tokens_count(encode_friendly(head))
      if last_paragraph is not None:
        last_paragraph.append(meta.page_index, head, tokens)
      else:
        last_paragraph = ParagraphDraft(meta.paragraph_type)
        last_paragraph.append(meta.page_index, head, tokens)

      for element in body:
        if last_paragraph is not None:
          yield last_paragraph
        tokens = self._llm.count_tokens_count(encode_friendly(element))
        last_paragraph = ParagraphDraft(meta.paragraph_type)
        last_paragraph.append(meta.page_index, element, tokens)

      if last_paragraph is not None:
        last_paragraph.set_tail_truncation(truncation)
        if truncation == TruncationKind.NO:
          yield last_paragraph
          last_paragraph = None

    if last_paragraph is not None:
      yield last_paragraph

  def _extract_sequences(self) -> Generator[tuple[int, Element], None, None]:
    for file_path, _, _, _ in xml_files(self._extraction_path):
      with open(file_path, mode="r", encoding="utf-8") as file:
        raw_page_xmls = fromstring(file.read())

      page_pairs: list[tuple[int, Element]] = []
      for page in raw_page_xmls:
        page_index = int(page.get("page-index", "-1"))
        page_pairs.append((page_index, page))

      page_pairs.sort(key=lambda x: x[0])
      for page_index, page in page_pairs:
        sequence = next(
          (found for found in page if found.get("type", None) == self._type),
          None
        )
        if sequence is None or len(sequence) == 0:
          continue

        paragraph_type = ParagraphType.TEXT
        if self._type == SequenceType.TEXT:
          paragraph_type = ParagraphType(page.get("type", "text"))

        sequence.set("type", paragraph_type.value)
        sequence.set("page-index", str(page_index))

        for line in sequence:
          if line.tag == "abandon":
            line.tag = "text"

        for i, layout in enumerate(sequence):
          layout.set("id", f"{page_index}/{i + 1}")

        yield page_index, sequence

  def _split_sequence(self, sequence: Element) -> tuple[Element, list[Element]]:
    head: Element | None = None
    body: list[Element] = []

    for i, element in enumerate(sequence):
      if i == 0:
        head = element
      else:
        body.append(element)

    return head, body



import sys

from pathlib import Path
from typing import Generator
from dataclasses import dataclass

from ..sequence import read_paragraphs
from ..data import Paragraph


@dataclass
class _Buffer:
  paragraph: Paragraph
  paragraph_index: tuple[int, int]
  begin_layout_index: tuple[int, int]
  end_layout_index: tuple[int, int]

  def include(self, layout_index: tuple[int, int]) -> bool:
    return self.begin_layout_index <= layout_index <= self.end_layout_index

class ParagraphsReader:
  def __init__(self, from_path: Path):
    self._gen: Generator[Paragraph, None, None] = read_paragraphs(from_path)
    self._buffer: _Buffer | None = None

  def read(self, layout_index: tuple[int, int]) -> Paragraph | None:
    if self._buffer is None:
      self._forward(layout_index)
    while True:
      if self._buffer is None:
        return None # read to the end
      if layout_index > self._buffer.end_layout_index:
        self._forward(layout_index)
      elif layout_index >= self._buffer.begin_layout_index:
        return self._buffer.paragraph
      else:
        return None

  def _forward(self, layout_index: tuple[int, int]) -> None:
    # Note: paragraph index and layout index are different, and you cannot
    #       assume that there is a certain correspondence between them.
    while True:
      paragraph = next(self._gen, None)
      if paragraph is None:
        self._buffer = None
        break

      if len(paragraph.layouts) == 0:
        continue

      begin_layout_index: tuple[int, int] = (sys.maxsize, sys.maxsize)
      end_layout_index: tuple[int, int] = (-1, -1)

      for layout in paragraph.layouts:
        index = (layout.page_index, layout.order_index)
        if index < begin_layout_index:
          begin_layout_index = index
        if index > end_layout_index:
          end_layout_index = index

      if layout_index <= end_layout_index:
        self._buffer = _Buffer(
          paragraph=paragraph,
          paragraph_index=(paragraph.page_index, paragraph.order_index),
          begin_layout_index=begin_layout_index,
          end_layout_index=end_layout_index,
        )
        break


from pathlib import Path
from typing import TypedDict
from strenum import StrEnum

from ..reporter import Reporter, AnalysingStep
from ..utils import Context
from .footnote import append_footnote_for_chapters, generate_footnote_references


class _Phase(StrEnum):
  GENERATE_FOOTNOTES = "generate_footnotes"
  APPEND_FOOTNOTES = "append_footnotes"
  COMPLETED = "completed"

class _State(TypedDict):
  phase: _Phase

def generate_chapters_with_footnotes(
      reporter: Reporter,
      chapter_path: Path,
      footnote_sequence_path: Path,
      workspace_path: Path,
    ) -> Path:

  output_path = workspace_path / "output"
  context: Context[_State] = Context(
    reporter=reporter,
    path=workspace_path,
    init=lambda: {
      "phase": _Phase.GENERATE_FOOTNOTES.value,
    },
  )
  if context.state["phase"] == _Phase.GENERATE_FOOTNOTES:
    reporter.go_to_step(AnalysingStep.GENERATE_FOOTNOTES)
    generate_footnote_references(
      sequence_path=footnote_sequence_path,
      output_path=context.path,
    )
    context.state = {
      **context.state,
      "phase": _Phase.APPEND_FOOTNOTES.value,
    }

  if context.state["phase"] == _Phase.APPEND_FOOTNOTES:
    append_footnote_for_chapters(
      chapter_path=chapter_path,
      footnote_path=context.path,
      output_path=output_path,
    )
    context.state = {
      **context.state,
      "phase": _Phase.COMPLETED.value,
    }
  return output_path



import os
import re

from pathlib import Path
from typing import Generator
from xml.etree.ElementTree import fromstring, Element

XML_Info = tuple[Path, str, int, int]


def remove_file(file_path: Path) -> None:
  if file_path.exists():
    os.unlink(file_path)

def read_xml_file(file_path: Path) -> Element:
  with file_path.open("r", encoding="utf-8") as file:
    return fromstring(file.read())

def xml_files(dir_path: Path) -> list[XML_Info]:
  xml_infos: list[XML_Info] = []
  for file in dir_path.iterdir():
    file_path = dir_path / file
    if not file_path.is_file():
      continue
    file_prefix, index1, index2 = _split_index_suffix(file.name)
    if file_prefix is None:
      continue

    xml_infos.append((file_path, file_prefix, index1, index2))

  xml_infos.sort(key=lambda x: (x[2], x[3]))
  return xml_infos

def search_xml_children(parent: Element) -> Generator[tuple[Element, Element], None, None]:
  for child in parent:
    yield child, parent
    yield from search_xml_children(child)

def _split_index_suffix(file_name: str) -> tuple[str | None, int, int]:
  matches = re.match(r"^[a-zA-Z]+_\d+(_\d+)?\.xml$", file_name)
  if not matches:
    return None, -1, -1

  file_prefix: str
  index1: str
  index2: str
  cells = re.sub(r"\..*$", "", file_name).split("_")
  if len(cells) == 3:
    file_prefix, index1, index2 = cells
  else:
    file_prefix, index1 = cells
    index2 = index1

  return file_prefix, int(index1), int(index2)


from __future__ import annotations
from dataclasses import dataclass
from strenum import StrEnum
from xml.etree.ElementTree import Element


@dataclass
class Paragraph:
  type: ParagraphType
  page_index: int
  order_index: int
  layouts: list[Layout]

  def to_xml(self) -> Element:
    element = Element("paragraph")
    element.set("type", self.type.value)
    for layout in self.layouts:
      element.append(layout.to_xml())
    return element

class ParagraphType(StrEnum):
  TEXT = "text"
  CONTENTS = "contents"
  REFERENCES = "references"
  COPYRIGHT = "copyright"

@dataclass
class Layout:
  kind: LayoutKind
  page_index: int
  order_index: int
  caption: Caption
  lines: list[Line]

  @property
  def id(self) -> str:
    return f"{self.page_index}/{self.order_index}"

  def to_xml(self) -> Element:
    element = Element(self.kind.value)
    element.set("id", self.id)
    for line in self.lines:
      element.append(line.to_xml())
    if len(self.caption.lines) > 0:
      element.append(self.caption.to_xml())
    return element

@dataclass
class AssetLayout(Layout):
  hash: bytes

  def to_xml(self) -> Element:
    element = super().to_xml()
    element.set("hash", self.hash.hex())
    return element

@dataclass
class FormulaLayout(AssetLayout):
  latex: str

  def to_xml(self) -> Element:
    element = super().to_xml()
    if self.latex:
      latex_element = Element("latex")
      latex_element.text = self.latex
      element.insert(0, latex_element)
    return element

class LayoutKind(StrEnum):
  TEXT = "text"
  HEADLINE = "headline"
  FIGURE = "figure"
  TABLE = "table"
  FORMULA = "formula"
  ABANDON = "abandon"

ASSET_LAYOUT_KINDS = (
  LayoutKind.FIGURE,
  LayoutKind.TABLE,
  LayoutKind.FORMULA,
)

@dataclass
class Caption:
  lines: list[Line]

  def to_xml(self) -> Element:
    element = Element("caption")
    for line in self.lines:
      element.append(line.to_xml())
    return element

@dataclass
class Line:
  text: str
  confidence: str

  def to_xml(self) -> Element:
    element = Element("line")
    element.text = self.text
    element.set("confidence", str(object=self.confidence))
    return element